from django import forms
from django.conf import settings
from django.core.mail import EmailMessage
from django.template import loader

from crispy_forms.bootstrap import StrictButton
from crispy_forms.helper import FormHelper
from crispy_forms.layout import Layout


class ContactForm(forms.Form):

    subject_template = "wagtail_extensions/email/subject.txt"
    txt_template = "wagtail_extensions/email/message.txt"

    name = forms.CharField(max_length=100, required=True)
    email = forms.EmailField(required=True)
    message = forms.CharField(
        widget=forms.Textarea(attrs={'rows': 4, 'cols': False}),
        required=True,
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.form_tag = False
        self.helper.layout = Layout(
            'name',
            'email',
            'message',
            StrictButton('Send', type="submit", css_class="btn btn-primary"),
        )

    def get_email_context(self):
        ctx = self.cleaned_data.copy()
        ctx['subject_prefix'] = settings.EMAIL_SUBJECT_PREFIX
        return ctx

    def send_email(self, to, subject_template, txt_template, reply_to):
        context = self.get_email_context()
        from_email = settings.DEFAULT_FROM_EMAIL
        subject = loader.render_to_string(subject_template, context)
        message = loader.render_to_string(txt_template, context)
        msg = EmailMessage(
            subject.strip(),
            message,
            from_email,
            to,
            reply_to=reply_to
        )
        msg.send()

    def get_to(self, page):
        if page.enquiry_email:
            return [page.enquiry_email]
        else:
            return [e for _, e in settings.MANAGERS]

    def save(self, page):
        # Forward enquiry
        to = self.get_to(page)
        reply_to = [self.cleaned_data['email']]
        self.send_email(to, self.subject_template, self.txt_template, reply_to)



import calendar
from collections import defaultdict
import datetime
from functools import partial
from itertools import groupby
import math
import uuid

from dateutil.relativedelta import relativedelta
from django.core.cache import cache
from django.core.exceptions import ValidationError
from django.forms.utils import ErrorList
from django.utils.functional import cached_property
from django.utils.timezone import now
from phonenumber_field import phonenumber
from phonenumber_field.formfields import PhoneNumberField
from wagtail.core import blocks
from wagtail.documents.blocks import DocumentChooserBlock
from wagtail.images.blocks import ImageChooserBlock
from wagtailgeowidget.blocks import GeoBlock

from . import app_settings
from . import utils


class StrippedListBlock(blocks.ListBlock):
    """
    Does not save empty values
    """
    def get_prep_value(self, value):
        return [
            self.child_block.get_prep_value(item)
            for item in value if item != ''
        ]


class LinkBlockStructValue(blocks.StructValue):

    @cached_property
    def link_url(self):
        try:
            link_item = self['link'][0]
        except IndexError:
            return ''

        if link_item.block_type in ('page', 'document'):
            return link_item.value.url if link_item.value else ''
        else:
            return link_item.value      # raw url

    @cached_property
    def link_text(self):
        try:
            link_item = self['link'][0]
        except IndexError:
            return ''

        if link_item.block_type in ('page', 'document'):
            link_text = link_item.value.title  if link_item.value else ''
        else:
            link_text = link_item.value

        # If text is set then it takes precedence
        if self.get('text'):
            link_text = self['text']

        return link_text


class LinkBlock(blocks.StructBlock):

    text = blocks.CharBlock(required=False)
    # required=False because of https://github.com/wagtail/wagtail/issues/2665
    link = blocks.StreamBlock([
        ('page', blocks.PageChooserBlock()),
        ('document', DocumentChooserBlock()),
        ('url', blocks.CharBlock(label="URL (absolute or relative)"))
    ], max_num=1, required=False)

    def clean(self, value):
        result = super().clean(value)
        if len(result['link']) == 0 and getattr(self.meta, 'required', True):
            errors = {
                'link': ErrorList([
                    blocks.StreamBlockValidationError(non_block_errors=ErrorList([
                        ValidationError('A page, document or URL must be defined.'),
                    ]))
                ]),
            }
            raise ValidationError('LinkBlock validation error', params=errors)
        return result


    class Meta:
        template = 'wagtail_extensions/blocks/link.html'
        value_class = LinkBlockStructValue


class CarouselItemBlock(blocks.StructBlock):

    image = ImageChooserBlock()
    caption = blocks.CharBlock(required=False)
    link = LinkBlock(required=False)

    class Meta:
        template = 'wagtail_extensions/blocks/carousel_item.html'


class CarouselBlock(blocks.StructBlock):

    items = blocks.ListBlock(CarouselItemBlock())
    show_thumbnails = blocks.BooleanBlock(default=False, required=False)

    class Meta:
        template = 'wagtail_extensions/blocks/carousel.html'

    def get_context(self, value, parent_context=None):
        ctx = super().get_context(value, parent_context=parent_context)
        ctx['dom_id'] = 'carousel-'+uuid.uuid4().hex
        ctx['show_indicators'] = len(value.get('items', [])) > 1
        return ctx


class AddressBlock(blocks.StructBlock):

    lines = blocks.ListBlock(blocks.CharBlock(label="Line", required=False))

    class Meta:
        template = 'wagtail_extensions/blocks/address.html'


class PhoneBlock(blocks.FieldBlock):

    def __init__(self, required=True, help_text=None, **kwargs):
        self.field = PhoneNumberField(required=required, help_text=help_text)
        super().__init__(**kwargs)

    def get_prep_value(self, value):
        return str(value)

    def to_python(self, value):
        return phonenumber.to_python(value)


class DepartmentBlock(blocks.StructBlock):

    name = blocks.CharBlock(required=False)
    phones = StrippedListBlock(PhoneBlock(required=False))
    email = blocks.EmailBlock(required=False)
    primary = blocks.BooleanBlock(required=False, default=False)

    def clean(self, value):
        phones = value.get('phones', [])
        email = value.get('email')
        if not phones and not email:
            errors = {
                'phone': ErrorList([
                    ValidationError('Either a phone or email must be defined'),
                ]),
            }
            raise ValidationError("There is a problem with this department", params=errors)
        return super().clean(value)


class OpeningTimeBlock(blocks.StructBlock):
    """
    A semi-structured opening times block.
    """
    PUBLIC = 7
    PUBLIC_LABEL = 'Public holiday'
    WEEKDAYS = tuple(enumerate(calendar.day_name))
    ALL_DAYS = WEEKDAYS + ((PUBLIC, PUBLIC_LABEL),)

    weekday = blocks.ChoiceBlock(choices=ALL_DAYS, required=False)
    label = blocks.CharBlock(help_text='Optionally override default weekday label', required=False)
    date = blocks.DateBlock(help_text='Optionally specify a day these times are for', required=False)
    start = blocks.TimeBlock(required=False)
    end = blocks.TimeBlock(required=False)
    closed = blocks.BooleanBlock(default=False, required=False)

    class Meta:
        template = 'wagtail_extensions/blocks/opening_time.html'

    def clean(self, value):
        cleaned = super().clean(value)
        start, end, weekday, date = map(cleaned.get, ['start', 'end', 'weekday', 'date'])
        errors = defaultdict(ErrorList)

        if date and date < datetime.date.today():
            err = ValidationError('Dates need to be in the future')
            errors['date'].append(err)

        if (start or end) and not (weekday or date):
            err = ValidationError('Specifying a weekday or date is required')
            errors['weekday'].append(err)

        if start and end and end <= start:
            err = ValidationError('End must be after start')
            errors['end'].append(err)

        if errors:
            raise ValidationError("There is a problem with this opening time", params=errors)

        return cleaned

    def to_python(self, value):
        value = super().to_python(value)
        weekday = value.get('weekday')
        if weekday is not None and weekday != '':
            value['weekday'] = int(weekday)

            label = value.get('label')
            if value['weekday'] == self.PUBLIC and (label is None or label == ''):
                value['label'] = self.PUBLIC_LABEL
        return value

    @classmethod
    def single_date(cls, value):
        if value.get('date'):
            return True
        elif value.get('weekday') == cls.PUBLIC:
            return True
        else:
            return False

    @classmethod
    def next_date(cls, value):
        weekday = value.get('weekday')
        if weekday is not None and weekday != cls.PUBLIC:
            # Next date with this weekday
            today = now().date()
            return today + relativedelta(weekday=weekday)
        else:
            return None


class OpeningTimesBlock(blocks.StructBlock):
    """
    Using a StructBlock as subclassing ListBlock leads to problems when
    Wagtail does template rendering.
    """
    times = blocks.ListBlock(OpeningTimeBlock(required=False))

    class Meta:
        template = 'wagtail_extensions/blocks/opening_times.html'

    @staticmethod
    def time_keyfunc(opening_time):
        if opening_time.block.single_date(opening_time):
            return opening_time
        else:
            return (
                opening_time.get('closed'),
                opening_time.get('start'),
                opening_time.get('end'),
            )

    @classmethod
    def group_times(cls, value):
        # Read out groups to simplify templates
        return [list(group) for key, group in groupby(value, cls.time_keyfunc)]

    def get_context(self, value, parent_context=None):
        ctx = super().get_context(value, parent_context=parent_context)
        ctx['times'] = self.group_times(value.get('times'))
        ctx['today'] = self.opening_today(value)
        return ctx

    @staticmethod
    def get_time_for_date(value, date):
        if value:
            times = value.get('times')
            specific_times = utils.first_true(times, lambda x: x.get('date') == date)
            times = specific_times or utils.first_true(times, lambda x: x.get('weekday') == date.weekday())
            if times:
                return dict(times)
        return None

    @classmethod
    def opening_today(cls, value, cache_key=None):
        today = now().date()
        partialed_getter = partial(cls.get_time_for_date, value, today)
        if cache_key:
            return cache.get_or_set(cache_key, partialed_getter, 60*60*24)
        else:
            return partialed_getter()


class LocationBlock(blocks.StructBlock):
    name = blocks.CharBlock()
    address = AddressBlock(required=False)
    point = GeoBlock(required=False)
    departments = blocks.ListBlock(DepartmentBlock(label="department", required=False))
    opening_times = OpeningTimesBlock(required=False)
    primary = blocks.BooleanBlock(default=False, required=False)


class SocialMediaProfileBlock(blocks.StructBlock):

    icon = blocks.ChoiceBlock(choices=app_settings.SOCIAL_MEDIA_TYPES)
    url = blocks.URLBlock()

    class Meta:
        template = 'wagtail_extensions/blocks/social_media_profile.html'


class TextBlock(blocks.StructBlock):

    title = blocks.CharBlock(required=False)
    body = blocks.RichTextBlock()

    class Meta:
        template = 'wagtail_extensions/blocks/text.html'


class ImagesBlock(blocks.StructBlock):

    images = blocks.ListBlock(ImageChooserBlock(required=False))

    class Meta:
        template = 'wagtail_extensions/blocks/images.html'

    def get_context(self, value, parent_context=None):
        ctx = super().get_context(value, parent_context=parent_context)
        images = value.get('images', None)
        ctx['column_width'] = math.floor(12 / len(images)) if images else 12
        return ctx



# Generated by Django 2.0.3 on 2018-04-13 11:32

from django.db import migrations, models
import django.db.models.deletion
import wagtail.core.blocks
import wagtail.core.fields
import wagtail_extensions.blocks
import wagtailgeowidget.blocks


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('wagtailcore', '0040_page_draft_title'),
    ]

    operations = [
        migrations.CreateModel(
            name='ContactDetailsTestSetting',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('locations', wagtail.core.fields.StreamField((('location', wagtail.core.blocks.StructBlock((('name', wagtail.core.blocks.CharBlock()), ('address', wagtail.core.blocks.StructBlock((('lines', wagtail.core.blocks.ListBlock(wagtail.core.blocks.CharBlock(label='Line', required=False))),), required=False)), ('point', wagtailgeowidget.blocks.GeoBlock(required=False)), ('departments', wagtail.core.blocks.ListBlock(wagtail.core.blocks.StructBlock((('name', wagtail.core.blocks.CharBlock(required=False)), ('phones', wagtail_extensions.blocks.StrippedListBlock(wagtail_extensions.blocks.PhoneBlock(required=False))), ('email', wagtail.core.blocks.EmailBlock(required=False)), ('primary', wagtail.core.blocks.BooleanBlock(default=False, required=False))), label='department', required=False))), ('opening_times', wagtail.core.blocks.StructBlock((('times', wagtail.core.blocks.ListBlock(wagtail.core.blocks.StructBlock((('weekday', wagtail.core.blocks.ChoiceBlock(choices=[(0, 'Monday'), (1, 'Tuesday'), (2, 'Wednesday'), (3, 'Thursday'), (4, 'Friday'), (5, 'Saturday'), (6, 'Sunday'), (7, 'Public holiday')], required=False)), ('label', wagtail.core.blocks.CharBlock(help_text='Optionally override default weekday label', required=False)), ('date', wagtail.core.blocks.DateBlock(help_text='Optionally specify a day these times are for', required=False)), ('start', wagtail.core.blocks.TimeBlock(required=False)), ('end', wagtail.core.blocks.TimeBlock(required=False)), ('closed', wagtail.core.blocks.BooleanBlock(default=False, required=False))), required=False))),), required=False)), ('primary', wagtail.core.blocks.BooleanBlock(default=False, required=False))))),))),
                ('site', models.OneToOneField(editable=False, on_delete=django.db.models.deletion.CASCADE, to='wagtailcore.Site')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



from django.contrib.postgres.fields import JSONField
from django.db import models
from django.utils.timezone import now

from wagtail.contrib.settings.models import BaseSetting
from wagtail.contrib.table_block.blocks import TableBlock
from wagtail.admin.edit_handlers import StreamFieldPanel
from wagtail.core import blocks, fields
from wagtail.core.models import Page
from wagtail.images.edit_handlers import ImageChooserPanel

from . import blocks as extension_blocks
from . import utils


class ContactSubmission(models.Model):

    date_submitted = models.DateTimeField(auto_now_add=True)
    data = JSONField()

    def __str__(self):
        if self.data.get('name'):
            return 'Submission from {} on {}'.format(self.data['name'], self.date_submitted)
        return 'Submission on {}'.format(self.date_submitted)

    class Meta:
        ordering = ['-date_submitted']


class ContentPage(Page):

    class Meta:
        abstract = True

    featured_image = models.ForeignKey(
        'wagtailimages.Image',
        models.PROTECT,
        null=True,
        blank=True,
        related_name='+'
    )
    body = fields.StreamField([
        ('text', extension_blocks.TextBlock()),
        ('table', TableBlock()),
        ('images', extension_blocks.ImagesBlock()),
        ('carousel', extension_blocks.CarouselBlock()),
    ], blank=True)

    content_panels = Page.content_panels + [
        ImageChooserPanel('featured_image'),
        StreamFieldPanel('body'),
    ]


class LinksSetting(BaseSetting):

    class Meta:
        abstract = True

    links = fields.StreamField([
        ('link', extension_blocks.LinkBlock()),
    ], null=True, blank=True)

    panels = (
        StreamFieldPanel('links'),
    )


class ContactDetailsSetting(BaseSetting):

    CACHE_KEY_OPENING_TODAY = "wagtail_extensions_opening_today_{:%Y%m%d}"

    locations = fields.StreamField([
        ('location', extension_blocks.LocationBlock()),
    ])

    panels = (
        StreamFieldPanel('locations'),
    )

    class Meta:
        abstract = True

    @classmethod
    def get_opening_today_cache_key(cls):
        today = now().date()
        return cls.CACHE_KEY_OPENING_TODAY.format(today)

    @property
    def primary_location(self):
        return utils.true_or_nth(self.locations, lambda x: x.value.get('primary') == True)

    @property
    def primary_department(self):
        location = self.primary_location
        if location:
            departments = location.value.get('departments', [])
            return utils.true_or_nth(departments, lambda x: x.get('primary') == True)
        return None

    @property
    def primary_phone(self):
        department = self.primary_department
        if department:
            phones = self.primary_department.get('phones', [])
            return utils.nth(phones, 0)
        return None

    @property
    def primary_opening_times(self):
        location = self.primary_location
        if location:
            return location.value.get('opening_times')
        return None

    @property
    def primary_opening_today(self):
        if self.primary_opening_times:
            cache_key = self.get_opening_today_cache_key()
            return self.primary_opening_times.block.opening_today(self.primary_opening_times, cache_key=cache_key)
        return None


class SocialMediaSetting(BaseSetting):

    class Meta:
        abstract = True

    profiles = fields.StreamField([
        ('profile', extension_blocks.SocialMediaProfileBlock()),
    ])

    panels = (
        StreamFieldPanel('profiles'),
    )



from django.contrib import messages
from django.db import models
from django.http import HttpResponseRedirect
from django.utils import timezone
from django.utils.decorators import method_decorator

from honeypot.decorators import check_honeypot
from wagtail.admin.edit_handlers import FieldPanel
from wagtail.core.models import Page

from .forms import ContactForm
from .models import ContactSubmission


class SiteSingleton:

    @classmethod
    def can_create_at(cls, parent):
        """
        Allows only one instance of this class to be created on a site.
        """
        can_create = super().can_create_at(parent)
        can_create_for_site = True
        site = parent.get_site()

        if site:
            can_create_for_site = not site.root_page.get_descendants().type(cls).exists()
        return can_create and can_create_for_site


class ContactMixin(models.Model):

    form_class = ContactForm
    success_url = None
    store_submissions = True
    success_message = 'Thank you! We will get back to you as soon as possible.'

    enquiry_email = models.EmailField(
        blank=True,
        help_text="Email address that will receive enquiries",
        null=True,
    )

    content_panels = Page.content_panels + [
        FieldPanel('enquiry_email'),
    ]

    class Meta:
        abstract = True

    def get_form(self, request):
        return self.form_class(request.POST or None)

    def store_submission(self, form_data):
        # We do this here, instead of in the form, so that a project
        # can make use of this regardless of which form it uses.
        if self.store_submissions:
            ContactSubmission.objects.create(data=form_data)

    @method_decorator(check_honeypot)
    def serve(self, request, *args, **kwargs):
        self.form = self.get_form(request)
        if request.method == 'POST':
            if self.form.is_valid():
                self.form.save(page=self)  # Save triggers an email
                self.store_submission(self.form.cleaned_data)
                # Add a message to be displayed to the user
                success_message = self.get_success_message()
                if success_message:
                    messages.add_message(request, messages.INFO, success_message)

                request.session['enquiry_form_submitted'] = timezone.now().strftime('%Y-%m-%d %H:%M %z')
                # Redirect to the current page, to prevent resubmissions
                return HttpResponseRedirect(self.get_success_url())

        return super().serve(request, *args, **kwargs)

    def get_context(self, request):
        ctx = super().get_context(request)
        ctx['form'] = self.form
        return ctx

    def get_success_url(self):
        return self.success_url or self.url

    def get_success_message(self):
        return self.success_message



from django.conf import settings


GEO_WIDGET_DEFAULT_LOCATION = getattr(settings, 'GEO_WIDGET_DEFAULT_LOCATION', {'lat': -1.3, 'lng': 36.8})
GEO_WIDGET_ZOOM = getattr(settings, 'GEO_WIDGET_ZOOM', 10)

SOCIAL_MEDIA_TYPES = getattr(settings, 'SOCIAL_MEDIA_TYPES', (
    ('facebook', 'Facebook'),
    ('twitter', 'Twitter'),
    ('instagram', 'Instagram'),
    ('linkedin', 'LinkedIn'),
))
